import collections
import contextlib
import functools
import inspect
import typing

import attrs
import typing_extensions

import interactions
from . import context
from . import converters
from . import errors
from .utils import _start_quotes
from .utils import maybe_coroutine

__all__ = (
    "MolterCommand",
    "prefixed_command",
    "prefix_command",
    "text_command",
    "text_based_command",
    "register_converter",
    "globally_register_converter",
    "prefixed_parameter",
    "ext_error_handler",
)


if typing.TYPE_CHECKING:
    from .base import MolterExtension

# 3.8+ compatibility
NoneType = type(None)

try:
    from types import UnionType  # type: ignore

    UNION_TYPES = {typing.Union, UnionType}
except ImportError:  # 3.8-3.9
    UNION_TYPES = {typing.Union}


# thankfully, modules are singletons
_global_type_to_converter = dict(converters.INTER_OBJECT_TO_CONVERTER)

ERROR_HANDLER = typing.Callable[
    [context.MolterContext, Exception], typing.Awaitable[typing.Any]
]
EXT_ERROR_HANDLER = typing.Callable[
    [typing.Any, context.MolterContext, Exception], typing.Awaitable[typing.Any]
]
# for some reason, typing an ext here doesn't work, so oh well
ERROR_HANDLER_FUNCTION = typing.Union[ERROR_HANDLER, EXT_ERROR_HANDLER]

EEH = typing.TypeVar("EEH", bound=EXT_ERROR_HANDLER)
EHF = typing.TypeVar("EHF", bound=ERROR_HANDLER_FUNCTION)


class ParameterInfo(typing.TypedDict):
    default: typing.Union[
        typing.Callable[[context.MolterContext, str], typing.Any], typing.Any
    ]


@attrs.define(slots=True)
class PrefixedCommandParameter:
    """
    An object representing parameters in a prefixed command.
    This class should not be instantiated directly.
    """

    name: str = attrs.field(default=None)
    "The name of the parameter."
    default: typing.Any = attrs.field(default=interactions.MISSING)
    "The default value of the parameter."
    type: typing.Type = attrs.field(default=None)
    "The type of the parameter."
    kind: inspect._ParameterKind = attrs.field(
        default=inspect.Parameter.POSITIONAL_OR_KEYWORD
    )
    """The kind of parameter this is as related to the function."""
    converters: typing.List[
        typing.Callable[[context.MolterContext, str], typing.Any]
    ] = attrs.field(factory=list)
    "A list of the converter functions for the parameter that convert to its type."
    greedy: bool = attrs.field(default=False)
    "Is the parameter greedy?"
    union: bool = attrs.field(default=False)
    "Is the parameter a union?"
    variable: bool = attrs.field(default=False)
    "Was the parameter marked as a variable argument?"
    consume_rest: bool = attrs.field(default=False)
    "Was the parameter marked to consume the rest of the input?"
    no_argument: bool = attrs.field(default=False)
    "Does this parameter have a converter that subclasses `NoArgumentConverter`?"

    @property
    def optional(self) -> bool:
        """Is this parameter optional?"""
        return self.default != interactions.MISSING


@attrs.define(slots=True)
class _PrefixedArgsIterator:
    """
    An iterator over the arguments of a prefixed command.
    Has functions to control the iteration.
    """

    args: typing.Tuple[str] = attrs.field()
    index: int = attrs.field(init=False, default=0)
    length: int = attrs.field(init=False, default=0)

    def __iter__(self) -> typing_extensions.Self:
        self.length = len(self.args)
        return self

    def __next__(self) -> str:
        if self.index >= self.length:
            raise StopIteration

        result = self.args[self.index]
        self.index += 1
        return self._remove_quotes(result)

    def _remove_quotes(self, arg: str) -> str:
        # this removes quotes from the arguments themselves
        return arg[1:-1] if arg[0] in _start_quotes else arg

    def _finish_args(self) -> typing.Tuple[str]:
        result = self.args[self.index - 1 :]
        self.index = self.length
        return result

    def get_rest_of_args(self) -> typing.Tuple[str]:
        return tuple(self._remove_quotes(r) for r in self._finish_args())

    def consume_rest(self) -> str:
        return " ".join(self._finish_args())

    def back(self, count: int = 1) -> None:
        self.index -= count

    def reset(self) -> None:
        self.index = 0

    @property
    def finished(self) -> bool:
        return self.index >= self.length


def _get_name(x: typing.Any):
    try:
        return x.__name__
    except AttributeError:
        return repr(x) if hasattr(x, "__origin__") else x.__class__.__name__


def _convert_to_bool(argument: str) -> bool:
    lowered = argument.lower()
    if lowered in {"yes", "y", "true", "t", "1", "enable", "on"}:
        return True
    elif lowered in {"no", "n", "false", "f", "0", "disable", "off"}:
        return False
    else:
        raise errors.BadArgument(f"{argument} is not a recognised boolean option.")


def _merge_converters(
    converter_dict: typing.Dict[type, typing.Type[converters.MolterConverter]]
) -> typing.Dict[type, typing.Type[converters.MolterConverter]]:
    # sourcery skip: dict-assign-update-to-union
    global _global_type_to_converter
    combined = dict(_global_type_to_converter)
    combined.update(converter_dict)
    return combined


def _get_from_anno_type(anno: typing_extensions.Annotated):
    """
    Handles dealing with Annotated annotations, getting their (first) type annotation.
    This allows correct type hinting with, say, Converters,
    for example.
    """
    # this is treated how it usually is during runtime
    # the first argument is ignored and the rest is treated as is
    args = typing_extensions.get_args(anno)[1:]
    return args[0]


def _get_converter_function(
    anno: typing.Union[
        typing.Type[converters.MolterConverter], converters.MolterConverter
    ],
    name: str,
) -> typing.Callable[[context.MolterContext, str], typing.Any]:
    num_params = len(inspect.signature(anno.convert).parameters.values())

    # if we have three parameters for the function, it's likely it has a self parameter
    # so we need to get rid of it by initing - typehinting hates this, btw!
    # the below line will error out if we aren't supposed to init it, so that works out
    actual_anno: converters.MolterConverter = anno() if num_params == 3 else anno  # type: ignore
    # we can only get to this point while having three params if we successfully inited
    if num_params == 3:
        num_params -= 1

    if num_params != 2:
        ValueError(
            f"{_get_name(anno)} for {name} is invalid: converters must have exactly 2"
            " arguments."
        )

    return actual_anno.convert


def _get_converter(
    anno: type,
    name: str,
    type_to_converter: typing.Dict[type, typing.Type[converters.MolterConverter]],
) -> typing.Callable[[context.MolterContext, str], typing.Any]:  # type: ignore
    if typing_extensions.get_origin(anno) == typing_extensions.Annotated:
        anno = _get_from_anno_type(anno)

    if isinstance(anno, converters.MolterConverter):
        return _get_converter_function(anno, name)

    elif converter := type_to_converter.get(anno, None):
        return _get_converter_function(converter, name)

    elif typing_extensions.get_origin(anno) is typing.Literal:
        literals = typing_extensions.get_args(anno)
        return converters._LiteralConverter(literals).convert

    elif inspect.isfunction(anno):
        num_params = len(inspect.signature(anno).parameters.values())
        if num_params == 0:
            ValueError(
                f"{_get_name(anno)} for {name} has 0 arguments, which is unsupported."
            )
        elif num_params == 1:
            return lambda ctx, arg: anno(arg)
        elif num_params == 2:
            return lambda ctx, arg: anno(ctx, arg)
        else:
            ValueError(
                f"{_get_name(anno)} for {name} has more than 2 arguments, which is"
                " unsupported."
            )

    elif anno == bool:
        return lambda ctx, arg: _convert_to_bool(arg)

    elif anno == inspect._empty:
        return lambda ctx, arg: str(arg)

    else:
        return lambda ctx, arg: anno(arg)


_INVALID_GREEDY_TYPES = {NoneType, str, converters.Greedy}.union(UNION_TYPES)


def _greedy_parse(greedy: converters.Greedy, param: inspect.Parameter):
    default = param.default

    if param.kind in {param.KEYWORD_ONLY, param.VAR_POSITIONAL}:
        raise ValueError("Greedy[...] cannot be a variable or keyword-only argument.")

    arg = typing_extensions.get_args(greedy)[0]

    if typing_extensions.get_origin(arg) == typing_extensions.Annotated:
        arg = _get_from_anno_type(arg)

    if typing.get_origin(arg) in UNION_TYPES:
        args = typing.get_args(arg)

        if len(args) > 2 or NoneType not in args:
            raise ValueError(f"Greedy[{repr(arg)}] is invalid.")

        arg = args[0]
        default = None

    if arg in _INVALID_GREEDY_TYPES:
        raise ValueError(f"Greedy[{_get_name(arg)}] is invalid.")

    return arg, default


def _get_params(
    signature: inspect.Signature,
    type_to_converter: typing.Dict[type, typing.Type[converters.MolterConverter]],
    parameter_info: typing.Dict[str, ParameterInfo],
):
    cmd_params: list[PrefixedCommandParameter] = []

    # this is used by keyword-only and variable args to make sure there isn't more than one of either
    # mind you, we also don't want one keyword-only and one variable arg either
    finished_params = False
    params = signature.parameters

    for name, param in params.items():
        if finished_params:
            raise ValueError("Cannot have multiple keyword-only or variable arguments.")

        cmd_param = PrefixedCommandParameter()
        cmd_param.name = name
        cmd_param.kind = param.kind
        cmd_param.type = anno = param.annotation

        if parameter_dict := parameter_info.pop(name, None):
            cmd_param.default = parameter_dict["default"]
        else:
            cmd_param.default = (
                param.default
                if param.default is not param.empty
                else interactions.MISSING
            )

        if typing_extensions.get_origin(anno) == converters.Greedy:
            anno, default = _greedy_parse(anno, param)

            if default is not param.empty:
                cmd_param.default = default

        if typing_extensions.get_origin(anno) in UNION_TYPES:
            cmd_param.union = True
            for arg in typing_extensions.get_args(anno):
                if isinstance(anno, converters.NoArgumentConverter) or issubclass(
                    type_to_converter.get(anno, object), converters.NoArgumentConverter
                ):
                    cmd_param.no_argument = True

                if arg != NoneType:
                    converter = _get_converter(arg, name, type_to_converter)
                    cmd_param.converters.append(converter)
                elif not cmd_param.optional:  # d.py-like behavior
                    cmd_param.default = None
        else:
            if isinstance(anno, converters.NoArgumentConverter) or issubclass(
                type_to_converter.get(anno, object), converters.NoArgumentConverter
            ):
                cmd_param.no_argument = True

            converter = _get_converter(anno, name, type_to_converter)  # type: ignore
            cmd_param.converters.append(converter)

        if param.kind == param.KEYWORD_ONLY:
            if cmd_param.greedy:
                raise ValueError("Keyword-only arguments cannot be Greedy.")

            cmd_param.consume_rest = True
            finished_params = True
        elif param.kind == param.VAR_POSITIONAL:
            if cmd_param.optional:
                # there's a lot of parser ambiguities here, so i'd rather not
                raise ValueError(
                    "Variable arguments cannot have default values or be Optional."
                )
            if cmd_param.greedy:
                raise ValueError("Variable arguments cannot be Greedy.")

            cmd_param.variable = True
            finished_params = True

        cmd_params.append(cmd_param)

    if parameter_info:
        for name in parameter_info.keys():
            # this should only raise once due to how errors work
            raise ValueError(f'"{name}" is not a parameter in the command provided.')

    return cmd_params


async def _convert(
    param: PrefixedCommandParameter, ctx: context.MolterContext, arg: str
):
    converted = interactions.MISSING
    for converter in param.converters:
        try:
            converted = await maybe_coroutine(converter, ctx, arg)
            break
        except Exception as e:
            if not param.union and not param.optional:
                if isinstance(e, errors.BadArgument):
                    raise
                raise errors.BadArgument(str(e)) from e

    used_default = False
    if converted == interactions.MISSING:
        if param.optional:
            converted = (
                await maybe_coroutine(param.default, ctx)
                if inspect.isfunction(param.default)
                else param.default
            )
            used_default = True
        else:
            union_types = typing_extensions.get_args(param.type)
            union_names = tuple(_get_name(t) for t in union_types)
            union_types_str = ", ".join(union_names[:-1]) + f", or {union_names[-1]}"
            raise errors.BadArgument(
                f'Could not convert "{arg}" into {union_types_str}.'
            )

    if param.no_argument:
        # tells converter not to eat the current argument
        used_default = True

    return converted, used_default


async def _greedy_convert(
    param: PrefixedCommandParameter,
    ctx: context.MolterContext,
    args: _PrefixedArgsIterator,
):
    args.back()
    broke_off = False
    greedy_args = []

    for arg in args:
        try:
            greedy_arg, used_default = await _convert(param, ctx, arg)

            if used_default:
                raise errors.BadArgument()  # does it matter?

            greedy_args.append(greedy_arg)
        except errors.BadArgument:
            broke_off = True
            break

    if not greedy_args:
        if param.optional:
            greedy_args = (
                await maybe_coroutine(param.default, ctx)
                if inspect.isfunction(param.default)
                else param.default
            )
        else:
            raise errors.BadArgument(
                f"Failed to find any arguments for {repr(param.type)}."
            )

    return greedy_args, broke_off


@attrs.define(
    slots=True,
    kw_only=True,
    hash=False,
)
class MolterCommand:
    extension: typing.Optional["MolterExtension"] = attrs.field(default=None)
    "The extension this command belongs to."
    enabled: bool = attrs.field(default=True)
    "Whether this can be run at all."
    callback: typing.Callable[..., typing.Coroutine] = attrs.field(
        default=None,
    )
    "The coroutine to be called for this command"
    name: str = attrs.field()
    "The name of the command."

    parameters: typing.List[PrefixedCommandParameter] = attrs.field(factory=list)
    "The paramters of the command."
    aliases: typing.List[str] = attrs.field(
        factory=list,
    )
    "The list of aliases the command can be invoked under."
    hidden: bool = attrs.field(
        default=False,
    )
    "If `True`, the default help command does not show this in the help output."
    ignore_extra: bool = attrs.field(default=True)
    """
    If `True`, ignores extraneous strings passed to a command if all its
    requirements are met (e.g. ?foo a b c when only expecting a and b).
    Otherwise, an error is raised. Defaults to True.
    """
    hierarchical_checking: bool = attrs.field(default=True)
    """If `True` and if the base of a subcommand, every subcommand underneath
    it will run this command's checks and cooldowns before its own. Otherwise,
    only the subcommand's checks are checked."""

    help: typing.Optional[str] = attrs.field(default=None)
    """The long help text for the command."""
    brief: typing.Optional[str] = attrs.field(default=None)
    "The short help text for the command."
    parent: typing.Optional[typing_extensions.Self] = attrs.field(
        default=None,
    )
    "The parent command, if applicable."
    subcommands: typing.Dict[str, typing_extensions.Self] = attrs.field(
        factory=dict,
    )
    "A dict of all subcommands for the command."

    checks: typing.List[
        typing.Callable[
            [context.MolterContext], typing.Coroutine[typing.Any, typing.Any, bool]
        ]
    ] = attrs.field(factory=list)
    """A list of checks for this command."""
    error_callback: typing.Optional[ERROR_HANDLER_FUNCTION] = attrs.field(default=None)
    """The callback to use if an error occurs."""

    _usage: typing.Optional[str] = attrs.field(default=None, repr=False)
    _type_to_converter: typing.Dict[
        type, typing.Type[converters.MolterConverter]
    ] = attrs.field(factory=dict, converter=_merge_converters, repr=False)
    _parameter_info: typing.Dict[str, ParameterInfo] = attrs.field(
        factory=dict, repr=False
    )
    _signature: typing.Optional[inspect.Signature] = attrs.field(
        default=None, repr=False
    )

    def __attrs_post_init__(self) -> None:
        # for some reason, i can't seem to pick this up anywhere else
        if hasattr(self.callback, "_parameter_info"):
            self._parameter_info = self.callback._parameter_info

        # we have to do this afterwards as these rely on the callback
        # and its own value, which is impossible to get with attrs
        # methods, i think

        if self.help:
            self.help = inspect.cleandoc(self.help)
        else:
            self.help = inspect.getdoc(self.callback)
            if isinstance(self.help, bytes):
                self.help = self.help.decode("utf-8")

        if self.brief is None:
            self.brief = self.help.splitlines()[0] if self.help is not None else None

        if hasattr(self.callback, "__checks__"):
            self.checks = self.callback.__checks__

    def __hash__(self):
        return id(self)

    @property
    def usage(self) -> str:
        """
        A string displaying how the command can be used.
        If no string is set, it will default to the command's signature.
        Useful for help commands.
        """
        return self._usage or self.signature

    @usage.setter
    def usage(self, usage: str) -> None:
        self._usage = usage

    @property
    def qualified_name(self) -> str:
        """Returns the full qualified name of this command."""
        name_deq = collections.deque()
        command = self

        while command.parent is not None:
            name_deq.appendleft(command.name)
            command = command.parent

        name_deq.appendleft(command.name)
        return " ".join(name_deq)

    @property
    def all_commands(self) -> typing.FrozenSet[typing_extensions.Self]:
        """Returns all unique subcommands underneath this command."""
        return frozenset(self.subcommands.values())

    @property
    def signature(self) -> str:
        """Returns a POSIX-like signature useful for help command output."""
        if not self.parameters:
            return ""

        results = []

        for param in self.parameters:
            anno = param.type
            name = param.name

            if typing_extensions.get_origin(anno) == typing_extensions.Annotated:
                # prefixed commands can only have two arguments in an annotation anyways
                anno = _get_from_anno_type(anno)

            if not param.greedy and param.union:
                union_args = typing_extensions.get_args(anno)
                if len(union_args) == 2 and param.optional:
                    anno = union_args[0]

            if typing_extensions.get_origin(anno) is typing.Literal:
                # it's better to list the values it can be than display the variable name itself
                name = "|".join(
                    f'"{v}"' if isinstance(v, str) else str(v)
                    for v in typing_extensions.get_args(anno)
                )

            # we need to do a lot of manipulations with the signature
            # string, so using a deque as a string builder makes sense for performance
            result_builder: typing.Deque[str] = collections.deque()

            if param.optional and param.default is not None:
                # it would be weird making it look like name=None
                result_builder.append(f"{name}={param.default}")
            else:
                result_builder.append(name)

            if param.variable:
                # this is inside the brackets
                result_builder.append("...")

            # surround the result with brackets
            if param.optional:
                result_builder.appendleft("[")
                result_builder.append("]")
            else:
                result_builder.appendleft("<")
                result_builder.append(">")

            if param.greedy:
                # this is outside the brackets, making it differentiable from
                # a variable argument
                result_builder.append("...")

            results.append("".join(result_builder))

        return " ".join(results)

    def _parse_parameters(self) -> None:
        """
        Parses the parameters that this command has into a form molter can use.

        This is purposely seperated like this to allow "lazy parsing" - parsing
        as the command is added to a client rather than being parsed immediately.
        This allows variables like "self" to be filtered out, and allows
        for decorators that modify parameters to be parsed later, all at once.

        This is called automatically if you add the command via
        `MolterManager.add_prefixed_command`, but it is up to the user to remember
        to put the extension in, if needed.
        """

        if not self._signature:
            # we want to avoid ctx - self should have been handled earlier
            callback = functools.partial(self.callback, None)
            self._signature = inspect.signature(callback)

        # doing this here just so we don't run into any issues here with a value
        # not being there yet or something if we used defaults, idk
        self.parameters = _get_params(
            self._signature,
            self._type_to_converter,
            self._parameter_info,
        )

        # we need to deal with subcommands too
        for command in self.all_commands:
            command._parse_parameters()

    def is_subcommand(self) -> bool:
        """Returns if this command is a subcommand or not."""
        return bool(self.parent)

    def add_command(self, cmd: typing_extensions.Self) -> None:
        """
        Adds a command as a subcommand to this command.

        Args:
            cmd (`MolterCommand`): The command to add
        """
        cmd.parent = self  # just so we know this is a subcommand

        if self.subcommands.get(cmd.name):
            raise ValueError(
                "Duplicate command! Multiple commands share the name/alias:"
                f" {self.qualified_name} {cmd.name}."
            )
        self.subcommands[cmd.name] = cmd

        for alias in cmd.aliases:
            if self.subcommands.get(alias):
                raise ValueError(
                    "Duplicate command! Multiple commands share the name/alias:"
                    f" {self.qualified_name} {cmd.name}."
                )
            self.subcommands[alias] = cmd

    def remove_command(self, name: str) -> None:
        """
        Removes a command as a subcommand from this command if it exists.
        If an alias is specified, only the alias will be removed.

        Args:
            name (`str`): The command to remove.
        """
        command = self.subcommands.pop(name, None)

        if command is None:
            return

        if name in command.aliases:
            command.aliases.remove(name)
            return

        for alias in command.aliases:
            self.subcommands.pop(alias, None)

    def get_command(self, name: str) -> typing.Optional[typing_extensions.Self]:
        """
        Gets a subcommand from this command. Can get subcommands of subcommands if needed.
        Args:
            name (`str`): The command to search for.
        Returns:
            `MolterCommand`: The command object, if found.
        """
        if " " not in name:
            return self.subcommands.get(name)

        names = name.split()
        if not names:
            return None

        cmd = self.subcommands.get(names[0])
        if not cmd or not cmd.subcommands:
            return cmd

        for name in names[1:]:
            try:
                cmd = cmd.subcommands[name]
            except (AttributeError, KeyError):
                return None

        return cmd

    @typing.overload
    def error(self) -> typing.Callable[[EHF], EHF]:
        ...

    @typing.overload
    def error(self, error_callback: EHF) -> EHF:
        ...

    def error(
        self,
        error_callback: typing.Optional[EHF] = None,
    ) -> typing.Union[typing.Callable[[EHF], EHF], EHF]:
        """
        A decorator to register an error callback for a command.

        Args:
            error_callback (`Callable`):
            The error callback to register. It must be an asynchronous function that
            takes in MolterContext and an exception.
        """

        def decorator(error_callback: EHF):
            self.error_callback = error_callback
            return error_callback

        if error_callback is not None:
            self.error_callback = error_callback
            return error_callback  # type: ignore

        return decorator

    def subcommand(
        self,
        name: typing.Optional[str] = None,
        *,
        aliases: typing.Optional[typing.List[str]] = None,
        help: typing.Optional[str] = None,
        brief: typing.Optional[str] = None,
        usage: typing.Optional[str] = None,
        enabled: bool = True,
        hidden: bool = False,
        ignore_extra: bool = True,
        hierarchical_checking: bool = True,
        type_to_converter: typing.Optional[
            typing.Dict[type, typing.Type[converters.MolterConverter]]
        ] = None,
    ) -> typing.Callable[..., typing_extensions.Self]:
        """
        A decorator to declare a subcommand for a molter prefixed command.

        Parameters:
            name (`str`, optional): The name of the command.
            Defaults to the name of the coroutine.

            aliases (`list[str]`, optional): The list of aliases the
            command can be invoked under.

            help (`str`, optional): The long help text for the command.
            Defaults to the docstring of the coroutine, if there is one.

            brief (`str`, optional): The short help text for the command.
            Defaults to the first line of the help text, if there is one.

            usage(`str`, optional): A string displaying how the command
            can be used. If no string is set, it will default to the
            command's signature. Useful for help commands.

            enabled (`bool`, optional): Whether this command can be run
            at all. Defaults to True.

            hidden (`bool`, optional): If `True`, the default help
            command (when it is added) does not show this in the help
            output. Defaults to False.

            ignore_extra (`bool`, optional): If `True`, ignores extraneous
            strings passed to a command if all its requirements are met
            (e.g. ?foo a b c when only expecting a and b).
            Otherwise, an error is raised. Defaults to True.

            hierarchical_checking (`bool`, optional): If `True` and if the
            base of a subcommand, every subcommand underneath it will run this
            command's checks before its own. Otherwise, only the subcommand's
            checks are checked. Defaults to True.

            type_to_converter (`dict[type, type[MolterConverter]]`, optional): A dict
            that associates converters for types. This allows you to use
            native type annotations without needing to use `typing.Annotated`.
            If this is not set, only interactions.py classes will be converted using
            built-in converters.

        Returns:
            `molter.MolterCommand`: The command object.
        """

        def wrapper(func):
            cmd = MolterCommand(  # type: ignore
                callback=func,
                name=name or func.__name__,
                aliases=aliases or [],
                help=help,
                brief=brief,
                usage=usage,  # type: ignore
                enabled=enabled,
                hidden=hidden,
                ignore_extra=ignore_extra,
                hierarchical_checking=hierarchical_checking,
                type_to_converter=type_to_converter  # type: ignore
                or getattr(func, "_type_to_converter", {}),
            )
            self.add_command(cmd)
            return cmd

        return wrapper

    async def _run_checks(self, ctx: context.MolterContext) -> None:
        for c in self.checks:
            try:
                if not await c(ctx):
                    raise errors.CheckFailure(ctx, check=c)
            except errors.CheckFailure as e:
                # pass in check function
                raise errors.CheckFailure(ctx, e.message, check=c)

    async def __call__(self, ctx: context.MolterContext) -> None:
        """
        Runs the command with checks.

        Args:
            ctx (`context.MolterContext`): The context to use for this command.
        """
        await self._run_checks(ctx)
        return await self.invoke(ctx)

    async def invoke(self, ctx: context.MolterContext) -> None:
        """
        Runs the callback of this command.

        Args:
            ctx (`context.MolterContext`): The context to use for this command.
        """
        # sourcery skip: remove-empty-nested-block, remove-redundant-if, remove-unnecessary-else
        callback = self.callback

        if len(self.parameters) == 0:
            return await callback(ctx)
        else:
            new_args: list[typing.Any] = []
            kwargs: dict[str, typing.Any] = {}
            args = _PrefixedArgsIterator(tuple(ctx.args))
            param_index = 0

            for arg in args:
                while param_index < len(self.parameters):
                    param = self.parameters[param_index]

                    if param.consume_rest:
                        arg = args.consume_rest()

                    if param.variable:
                        args_to_convert = args.get_rest_of_args()
                        new_arg = [
                            await _convert(param, ctx, arg) for arg in args_to_convert
                        ]
                        new_arg = tuple(arg[0] for arg in new_arg)
                        new_args.extend(new_arg)
                        param_index += 1
                        break

                    if param.greedy:
                        greedy_args, broke_off = await _greedy_convert(param, ctx, args)

                        new_args.append(greedy_args)
                        param_index += 1
                        if broke_off:
                            args.back()

                        if param.default:
                            continue
                        else:
                            break

                    converted, used_default = await _convert(param, ctx, arg)
                    if param.kind in {
                        inspect.Parameter.POSITIONAL_ONLY,
                        inspect.Parameter.VAR_POSITIONAL,
                    }:
                        new_args.append(converted)
                    else:
                        kwargs[param.name] = converted
                    param_index += 1

                    if not used_default:
                        break

            if param_index < len(self.parameters):
                for param in self.parameters[param_index:]:
                    if param.no_argument:
                        converted, _ = await _convert(param, ctx, None)  # type: ignore
                        if param.kind in {
                            inspect.Parameter.POSITIONAL_ONLY,
                            inspect.Parameter.VAR_POSITIONAL,
                        }:
                            new_args.append(converted)
                        else:
                            kwargs[param.name] = converted
                            break
                        continue

                    if not param.optional:
                        raise errors.BadArgument(
                            f"{param.name} is a required argument that is missing."
                        )
                    else:
                        default = (
                            await maybe_coroutine(param.default, ctx)
                            if inspect.isfunction(param.default)
                            else param.default
                        )

                        if param.kind in {
                            inspect.Parameter.POSITIONAL_ONLY,
                            inspect.Parameter.VAR_POSITIONAL,
                        }:
                            new_args.append(default)
                        else:
                            kwargs[param.name] = default
                            break
            elif not self.ignore_extra and not args.finished:
                raise errors.BadArgument(f"Too many arguments passed to {self.name}.")

            return await callback(ctx, *new_args, **kwargs)


def prefixed_command(
    name: typing.Optional[str] = None,
    *,
    aliases: typing.Optional[typing.List[str]] = None,
    help: typing.Optional[str] = None,
    brief: typing.Optional[str] = None,
    usage: typing.Optional[str] = None,
    enabled: bool = True,
    hidden: bool = False,
    ignore_extra: bool = True,
    hierarchical_checking: bool = True,
    type_to_converter: typing.Optional[
        typing.Dict[type, typing.Type[converters.MolterConverter]]
    ] = None,
) -> typing.Callable[..., MolterCommand]:
    """
    A decorator to declare a coroutine as a molter prefixed command.

    Parameters:
        name (`str`, optional): The name of the command.
        Defaults to the name of the coroutine.

        aliases (`list[str]`, optional): The list of aliases the
        command can be invoked under.

        help (`str`, optional): The long help text for the command.
        Defaults to the docstring of the coroutine, if there is one.

        brief (`str`, optional): The short help text for the command.
        Defaults to the first line of the help text, if there is one.

        usage(`str`, optional): A string displaying how the command
        can be used. If no string is set, it will default to the
        command's signature. Useful for help commands.

        enabled (`bool`, optional): Whether this command can be run
        at all. Defaults to True.

        hidden (`bool`, optional): If `True`, the default help
        command (when it is added) does not show this in the help
        output. Defaults to False.

        ignore_extra (`bool`, optional): If `True`, ignores extraneous
        strings passed to a command if all its requirements are met
        (e.g. ?foo a b c when only expecting a and b).
        Otherwise, an error is raised. Defaults to True.

        hierarchical_checking (`bool`, optional): If `True` and if the
        base of a subcommand, every subcommand underneath it will run this
        command's checks before its own. Otherwise, only the subcommand's
        checks are checked. Defaults to True.

        type_to_converter (`dict[type, type[MolterConverter]]`, optional): A dict
        that associates converters for types. This allows you to use
        native type annotations without needing to use `typing.Annotated`.
        If this is not set, only interactions.py classes will be converted using
        built-in converters.

    Returns:
        `molter.MolterCommand`: The command object.
    """

    def wrapper(func):
        return MolterCommand(  # type: ignore
            callback=func,
            name=name or func.__name__,
            aliases=aliases or [],
            help=help,
            brief=brief,
            usage=usage,  # type: ignore
            enabled=enabled,
            hidden=hidden,
            ignore_extra=ignore_extra,
            hierarchical_checking=hierarchical_checking,
            type_to_converter=type_to_converter  # type: ignore
            or getattr(func, "_type_to_converter", {}),
        )

    return wrapper


prefix_command = prefixed_command
text_command = prefixed_command
text_based_command = prefixed_command


# molter command typevar - can be the function or the command
MCT = typing.TypeVar("MCT", typing.Callable, MolterCommand)


def register_converter(
    type_: type, converter: typing.Type[converters.MolterConverter]
) -> typing.Callable[..., MCT]:
    """
    A decorator that allows you to register converters for a type for a specific command.
    This allows for native type annotations without needing to use `typing(_extensions).Annotated`.
    Args:
        type_ (`type`): The type to register for.
        converter (`type[MolterConverter]`): The converter to use for the type.
    Returns:
        `Callable | MolterCommand`: Either the callback or the command.
        If this is used after using the `molter.prefixed_command` decorator, it will be a command.
        Otherwise, it will be a callback.
    """
    if not isinstance(converter, converters.MolterConverter):
        raise ValueError('Converter provided did not have a "convert" method.')

    def wrapper(command: MCT) -> MCT:
        if hasattr(command, "_type_to_converter"):
            command._type_to_converter[type_] = converter
        else:
            command._type_to_converter = {type_: converter}

        return command

    return wrapper


def globally_register_converter(
    anno_type: type, converter: typing.Type[converters.MolterConverter]
) -> None:
    """
    A decorator that allows you to register converters for commands decorated/made after this is run.
    This allows for native type annotations without needing to use `typing.Annotated`.
    Note that this does not retroactively register converters for commands already made.
    Args:
        anno_type (`type`): The type to register for.
        converter (`type[MolterConverter]`): The converter to use for the type.
    """
    if not isinstance(converter, converters.MolterConverter):
        raise ValueError('Converter provided did not have a "convert" method.')

    # hate me, but i think it makes sense here
    global _global_type_to_converter
    _global_type_to_converter.update({anno_type: converter})


def prefixed_parameter(
    parameter_name: str,
    *,
    default: typing.Any = interactions.MISSING,
    factory: typing.Optional[
        typing.Callable[[context.MolterContext], typing.Any]
    ] = None,
) -> typing.Callable[..., MCT]:
    """
    A decorator that allows you to adjust a parameter in a prefixed comamnd.

    Args:
        parameter_name (`str`): The parameter in the command to edit.
        default (`Any`, optional): The default value the parameter should have.
        factory (`Callable[MolterContext]`, optional): A factory function to use
        if no argument is provided for this parameter. The function can be asynchronous
        or synchronous, and should take in solely a `MolterContext`.

    Returns:
        `Callable | MolterCommand`: Either the callback or the command.
        If this is used after using the `molter.prefixed_command` decorator, it will be a command.
        Otherwise, it will be a callback.
    """

    def wrapper(command: MCT) -> MCT:
        parameter_info: ParameterInfo = {}  # type: ignore

        if factory is not None and default is not interactions.MISSING:
            raise ValueError("You cannot provide both a default and a factory.")
        elif factory is None and default is interactions.MISSING:
            raise ValueError("You must provide either a default or a factory.")

        if factory is not None:
            parameter_info["default"] = factory
        elif default is not interactions.MISSING:
            parameter_info["default"] = default

        if hasattr(command, "_parameters_info"):
            command._parameter_info[parameter_name] = parameter_info  # type: ignore
        else:
            command._parameter_info = {parameter_name: parameter_info}  # type: ignore

        return command

    return wrapper


@typing.overload
def ext_error_handler() -> typing.Callable[[EEH], EEH]:
    ...


@typing.overload
def ext_error_handler(error_callback: EEH) -> EEH:
    ...


def ext_error_handler(
    error_callback: typing.Optional[EEH] = None,
) -> typing.Union[typing.Callable[[EEH], EEH], EEH]:
    """
    A decorator to register an error callback for a MolterExtension.

    Args:
        error_callback (`Callable`):
        The error callback to register. It must be an asynchronous function that
        takes in `self`, MolterContext and an exception.
    """

    def decorator(error_callback: EEH):
        error_callback.__ext_molter_error__ = True
        return error_callback

    if error_callback is not None:
        error_callback.__ext_molter_error__ = True
        return error_callback

    return decorator



import re
import typing

import interactions
import interactions.api.error as inter_errors
from . import errors
from . import utils
from .context import MolterContext
from .utils import _wrap_lib_exception

__all__ = (
    "MolterConverter",
    "NoArgumentConverter",
    "IDConverter",
    "SnowflakeConverter",
    "MemberConverter",
    "UserConverter",
    "ChannelConverter",
    "RoleConverter",
    "GuildConverter",
    "MessageConverter",
    "Greedy",
    "INTER_OBJECT_TO_CONVERTER",
)


T = typing.TypeVar("T")
T_co = typing.TypeVar("T_co", covariant=True)


@typing.runtime_checkable
class MolterConverter(typing.Protocol[T_co]):
    async def convert(self, ctx: MolterContext, argument: str) -> T_co:
        raise NotImplementedError("Derived classes need to implement this.")


class NoArgumentConverter(typing.Generic[T_co]):
    """
    An indicator class for special type of converters that only uses the context.
    Arguments will be "eaten up" by converters otherwise.
    """

    async def convert(self, ctx: MolterContext, argument: str) -> T_co:
        raise NotImplementedError("Derived classes need to implement this.")


class _LiteralConverter(MolterConverter):
    values: typing.Dict

    def __init__(self, args: typing.Any):
        self.values = {arg: type(arg) for arg in args}

    async def convert(self, ctx: MolterContext, argument: str):
        for arg, converter in self.values.items():
            try:
                if (converted := converter(argument)) == arg:
                    return converted
            except Exception:
                continue

        literals_list = [str(a) for a in self.values.keys()]
        literals_str = ", ".join(literals_list[:-1]) + f", or {literals_list[-1]}"
        raise errors.BadArgument(
            f'Could not convert "{argument}" into one of {literals_str}.'
        )


_ID_REGEX = re.compile(r"([0-9]{15,})$")


class IDConverter(MolterConverter[T_co]):
    @staticmethod
    def _get_id_match(argument):
        return _ID_REGEX.match(argument)


class SnowflakeConverter(IDConverter[interactions.Snowflake]):
    async def convert(self, _, argument: str) -> interactions.Snowflake:
        match = self._get_id_match(argument) or re.match(
            r"<(?:@(?:!|&)?|#)([0-9]{15,})>$", argument
        )

        if match is None:
            raise errors.BadArgument(f'"{argument}" is not a valid snowflake.')

        return interactions.Snowflake(match.group(1))


class MemberConverter(IDConverter[interactions.Member]):
    def _get_member_from_list(self, members_data: typing.List[dict], argument: str):
        # sourcery skip: assign-if-exp
        result = None
        if len(argument) > 5 and argument[-5] == "#":
            result = next(
                (
                    m
                    for m in members_data
                    if f"{m['user']['username']}#{m['user']['discriminator']}"
                    == argument
                ),
                None,
            )

        if not result:
            result = next(
                (
                    m
                    for m in members_data
                    if m.get("nick") == argument or m["user"]["username"] == argument
                ),
                None,
            )

        if result is not None:
            return interactions.Member(**result)
        return result

    async def convert(self, ctx: MolterContext, argument: str) -> interactions.Member:
        if not ctx.guild or not ctx.guild_id:
            raise errors.BadArgument("This command cannot be used in private messages.")

        match = self._get_id_match(argument) or re.match(
            r"<@!?([0-9]{15,})>$", argument
        )
        result = None

        if match:
            result = await _wrap_lib_exception(
                interactions.get(
                    ctx.client,
                    interactions.Member,
                    parent_id=int(ctx.guild_id),
                    object_id=int(match.group(1)),
                )
            )
        else:
            query = argument
            if len(argument) > 5 and argument[-5] == "#":
                query, _, _ = argument.rpartition("#")

            members_data = await _wrap_lib_exception(
                ctx._http.search_guild_members(int(ctx.guild_id), query, limit=5)
            )

            if not members_data:
                raise errors.BadArgument(f'Member "{argument}" not found.')

            result = self._get_member_from_list(members_data, argument)

        if not result:
            raise errors.BadArgument(f'Member "{argument}" not found.')

        result._client = ctx._http
        return result


class UserConverter(IDConverter[interactions.User]):
    async def convert(self, ctx: MolterContext, argument: str) -> interactions.User:
        # sourcery skip: remove-redundant-pass
        match = self._get_id_match(argument) or re.match(
            r"<@!?([0-9]{15,})>$", argument
        )
        result = None

        if match:
            result = await _wrap_lib_exception(
                interactions.get(
                    ctx.client, interactions.User, object_id=int(match.group(1))
                )
            )
        else:
            all_cached_users = ctx._http.cache[interactions.User].values.values()
            if len(argument) > 5 and argument[-5] == "#":
                result = next(
                    (
                        u
                        for u in all_cached_users
                        if f"{u.username}#{u.discriminator}" == argument
                    ),
                    None,
                )

            if not result:
                result = next(
                    (u for u in all_cached_users if u.username == argument), None
                )

        if not result:
            raise errors.BadArgument(f'User "{argument}" not found.')

        result._client = ctx._http
        return result


class ChannelConverter(IDConverter[interactions.Channel]):
    async def convert(
        self,
        ctx: MolterContext,
        argument: str,
    ) -> interactions.Channel:
        match = self._get_id_match(argument) or re.match(r"<#([0-9]{15,})>$", argument)
        result = None

        if match:
            result = await _wrap_lib_exception(
                interactions.get(
                    ctx.client, interactions.Channel, object_id=int(match.group(1))
                )
            )
        elif ctx.guild and ctx.guild.channels:
            result = next(
                (
                    c
                    for c in ctx.guild.channels
                    if c.name == utils.remove_prefix(argument, "#")
                ),
                None,
            )

        if not result:
            raise errors.BadArgument(f'Channel "{argument}" not found.')

        result._client = ctx._http
        return result


class RoleConverter(IDConverter[interactions.Role]):
    async def convert(
        self,
        ctx: MolterContext,
        argument: str,
    ) -> interactions.Role:
        if not ctx.guild_id or not ctx.guild:
            raise errors.BadArgument("This command cannot be used in private messages.")

        # while i'd like to use the cache here, roles don't store info about which guild they come
        # from, and the guild doesn't update its role list
        match = self._get_id_match(argument) or re.match(r"<@&([0-9]{15,})>$", argument)
        result = None

        if match:
            result = await _wrap_lib_exception(
                interactions.get(
                    ctx.client,
                    interactions.Role,
                    parent_id=int(ctx.guild_id),
                    object_id=int(match.group(1)),
                )
            )
        else:
            result = next(
                (r for r in ctx.guild.roles if r.name == argument),
                None,
            )

        if not result:
            raise errors.BadArgument(f'Role "{argument}" not found.')

        return interactions.Role(**result, _client=ctx._http)  # type: ignore


class GuildConverter(IDConverter[interactions.Guild]):
    async def convert(self, ctx: MolterContext, argument: str) -> interactions.Guild:
        if match := self._get_id_match(argument):
            result = await _wrap_lib_exception(
                interactions.get(
                    ctx.client, interactions.Guild, object_id=int(match.group(1))
                )
            )
        else:
            result = next((g for g in ctx.bot.guilds if g.name == argument), None)

        if not result:
            raise errors.BadArgument(f'Guild "{argument}" not found.')

        result._client = ctx._http
        return result


class MessageConverter(MolterConverter[interactions.Message]):
    # either just the id or <chan_id>-<mes_id>, a format you can get by shift clicking "copy id"
    _ID_REGEX = re.compile(
        r"(?:(?P<channel_id>[0-9]{15,})-)?(?P<message_id>[0-9]{15,})"
    )
    # of course, having a way to get it from a link is nice
    _MESSAGE_LINK_REGEX = re.compile(
        r"https?://[\S]*?discord(?:app)?\.com/channels/(?P<guild_id>[0-9]{15,}|@me)/"
        r"(?P<channel_id>[0-9]{15,})/(?P<message_id>[0-9]{15,})\/?$"
    )

    async def convert(self, ctx: MolterContext, argument: str) -> interactions.Message:
        match = self._ID_REGEX.match(argument) or self._MESSAGE_LINK_REGEX.match(
            argument
        )
        if not match:
            raise errors.BadArgument(f'Message "{argument}" not found.')

        data = match.groupdict()

        message_id = int(data["message_id"])
        channel_id = (
            int(data["channel_id"]) if data.get("channel_id") else int(ctx.channel_id)
        )

        # this guild checking is technically unnecessary, but we do it just in case
        # it means a user cant just provide an invalid guild id and still get a message
        guild_id = data["guild_id"] if data.get("guild_id") else ctx.guild_id
        guild_id = str(guild_id) if guild_id != "@me" else None

        if cached_message := ctx._http.cache[interactions.Message].get(
            interactions.Snowflake(message_id)
        ):
            if int(cached_message.channel_id) == channel_id and (
                cached_message.guild_id == guild_id
                or str(cached_message.guild_id == guild_id)
            ):
                return cached_message
            else:
                raise errors.BadArgument(f'Message "{argument}" not found.')

        try:
            message_data: dict = await ctx._http.get_message(channel_id, message_id)  # type: ignore

            msg_guild_id: typing.Optional[str] = message_data.get("guild_id")
            if not msg_guild_id:
                if channel := ctx._http.cache[interactions.Channel].get(
                    interactions.Snowflake(channel_id)
                ):
                    msg_guild_id = message_data["guild_id"] = (
                        str(channel.guild_id) if channel.guild_id else None
                    )
                else:
                    channel_data = await _wrap_lib_exception(
                        ctx._http.get_channel(channel_id)
                    )
                    if channel_data:
                        msg_guild_id = message_data["guild_id"] = channel_data.get(
                            "guild_id"
                        )
                    else:
                        raise errors.BadArgument(f'Message "{argument}" not found.')

            if msg_guild_id != guild_id:
                raise errors.BadArgument(f'Message "{argument}" not found.')

            return interactions.Message(**message_data, _client=ctx._http)  # type: ignore
        except inter_errors.LibraryException:
            raise errors.BadArgument(f'Message "{argument}" not found.')


class AttachmentConverter(NoArgumentConverter[interactions.Attachment]):
    async def convert(self, ctx: MolterContext, _) -> interactions.Attachment:
        # could be edited by a dev, but... why
        attachment_counter: int = ctx.extras.get("__molter_attachment_counter", 0)

        try:
            attach = ctx.message.attachments[attachment_counter]
            ctx.extras["__molter_attachment_counter"] = attachment_counter + 1
            return attach
        except IndexError:
            raise errors.BadArgument("There are no more attachments for this context.")


class Greedy(typing.List[T]):
    # this class doesn't actually do a whole lot
    # it's more or less simply a note to the parameter
    # getter
    pass


INTER_OBJECT_TO_CONVERTER: typing.Dict[type, typing.Type[MolterConverter]] = {
    interactions.Snowflake: SnowflakeConverter,
    interactions.Member: MemberConverter,
    interactions.User: UserConverter,
    interactions.Channel: ChannelConverter,
    interactions.Role: RoleConverter,
    interactions.Guild: GuildConverter,
    interactions.Message: MessageConverter,
    interactions.Attachment: AttachmentConverter,  # type: ignore
}



import typing

from .utils import escape_mentions

if typing.TYPE_CHECKING:
    from .context import MolterContext


__all__ = ("MolterException", "BadArgument", "CheckFailure")


class MolterException(Exception):
    """The base exception for molter exceptions."""

    pass


class BadArgument(MolterException):
    """A special exception for invalid arguments when using molter commands."""

    def __init__(self, message: typing.Optional[str] = None, *args: typing.Any) -> None:
        if message is not None:
            message = escape_mentions(message)
            super().__init__(message, *args)
        else:
            super().__init__(*args)


class CheckFailure(MolterException):
    """
    An exception when a check fails.

    Attributes:
        context (`MolterContext`): The context for this check.

        message: (`str`, optional): The error message.

        check (`Callable[[MolterContext], typing.Coroutine[Any, Any, bool]]`, optional):
        The check that failed. This is automatically passed in if the check fails -
        there is no need to do it yourself.
    """

    def __init__(
        self,
        context: "MolterContext",
        message: typing.Optional[str] = "A check has failed.",
        *,
        check: typing.Callable[
            ["MolterContext"], typing.Coroutine[typing.Any, typing.Any, bool]
        ] = None,
    ):
        self.context = context
        self.check = check
        self.message = message

        super().__init__(message)



import inspect
import logging
import sys
import traceback
import typing

import interactions
from . import utils
from .command import MolterCommand
from .context import MolterContext
from .converters import MolterConverter
from interactions import ext

__all__ = (
    "__version__",
    "base",
    "MolterInjectedClient",
    "MolterExtensionMixin",
    "MolterExtension",
    "MolterManager",
    "setup",
)

__version__ = "0.6.1"

logger: logging.Logger = logging.getLogger("molter")


version = ext.Version(
    version=__version__,
    authors=[ext.VersionAuthor("Astrea49")],
)

base = ext.Base(
    name="interactions-molter",
    version=version,
    link="https://github.com/interactions-py/molter/",
    description=(
        "An extension library for interactions.py to add prefixed commands. A"
        " demonstration of molter-core."
    ),
    packages=["interactions.ext.molter"],
    requirements=["discord-py-interactions>=4.3.0"],
)


class MolterInjectedClient(interactions.Client):
    """
    A semi-stub for Clients injected with molter.
    This should only be used for typehinting.
    """

    molter: "MolterManager"


class MolterExtensionMixin:
    """
    A mixin that can be used to add molter functionality into any subclass of
    interactions.py's extensions. Simply use it like so:

    ```python
    class MyExt(MolterExtensionMixin, ActualExtension):
        ...
    ```
    """

    client: interactions.Client
    _error_callback: typing.Optional[
        typing.Callable[[MolterContext, Exception], typing.Coroutine]
    ] = None
    _molter_prefixed_commands: typing.List[MolterCommand]

    def __new__(cls, client: interactions.Client, *args, **kwargs):
        self = super().__new__(cls, client, *args, **kwargs)  # type: ignore
        self._molter_prefixed_commands = []

        # typehinting funkyness for better typehints
        self.client = typing.cast(MolterInjectedClient, self.client)

        error_handler_count = 0

        for _, func in inspect.getmembers(
            self,
            predicate=lambda x: isinstance(x, MolterCommand)
            or hasattr(x, "__ext_molter_error__"),
        ):
            if isinstance(func, MolterCommand):
                cmd: MolterCommand = func

                if not cmd.is_subcommand():  # we don't want to add subcommands
                    cmd = utils._wrap_recursive(cmd, self)
                    self._molter_prefixed_commands.append(cmd)
                    self.client.molter.add_prefixed_command(cmd)
            elif hasattr(func, "__ext_molter_error__"):
                if error_handler_count >= 1:
                    raise ValueError(
                        "A molter extension cannot have more than one molter command"
                        " error handler."
                    )

                self._error_callback = func
                error_handler_count += 1

        return self

    async def teardown(self, *args, **kwargs) -> None:
        # typehinting funkyness for better typehints
        self.client = typing.cast(MolterInjectedClient, self.client)

        for cmd in self._molter_prefixed_commands:
            names_to_remove = cmd.aliases.copy()
            names_to_remove.append(cmd.name)

            for name in names_to_remove:
                self.client.molter.prefixed_commands.pop(name, None)

        return await super().teardown(*args, **kwargs)  # type: ignore


class MolterExtension(MolterExtensionMixin, interactions.Extension):
    """An extension that allows you to use molter commands in them."""

    pass


class MolterManager:
    """
    The main part of the extension. Deals with injecting itself in the first place.

    Parameters:
        client (`interactions.Client`): The client instance.
        default_prefix (`str | typing.Iterable[str]`, optional): \
            The default prefix to use. Defaults to None.
        generate_prefixes (`typing.Callable`, optional): An asynchronous function \
            that takes in a `Client` and `Message` object and returns either a \
            string or an iterable of strings. Defaults to None.
        on_molter_command_error (`typing.Callable`, optional): An asynchronous function \
            that takes in a `MolterContext` and `Exception` to handle errors that occur \
            when running molter commands. By default, molter will output the error to \
            the default logging place and ignore it. The error event can also be listened \
            to by listening to the "on_molter_command_error" event.

        If neither `default_prefix` or `generate_prefixes` are provided, the bot
        defaults to using it being mentioned as its prefix.
    """

    def __init__(
        self,
        client: interactions.Client,
        *,
        default_prefix: typing.Optional[typing.Union[str, typing.Iterable[str]]] = None,
        generate_prefixes: typing.Optional[
            typing.Callable[
                [interactions.Client, interactions.Message],
                typing.Coroutine[
                    typing.Any, typing.Any, typing.Union[str, typing.Iterable[str]]
                ],
            ]
        ] = None,
        on_molter_command_error: typing.Optional[
            typing.Callable[[MolterContext, Exception], typing.Coroutine]
        ] = None,
    ) -> None:
        # typehinting funkyness for better typehints
        client = typing.cast(MolterInjectedClient, client)

        self.client = client
        self.default_prefix = default_prefix
        self.prefixed_commands: typing.Dict[str, MolterCommand] = {}

        if default_prefix is None and generate_prefixes is None:
            # by default, use mentioning the bot as the prefix
            generate_prefixes = utils.when_mentioned

        self.generate_prefixes = (  # type: ignore
            generate_prefixes
            if generate_prefixes is not None
            else self.generate_prefixes
        )
        self.on_molter_command_error = (  # type: ignore
            on_molter_command_error
            if on_molter_command_error is not None
            else self.on_molter_command_error
        )

        # this allows us to use a (hopefully) non-conflicting namespace
        self.client.molter = self

        # i hope someone dies internally when looking at this /lhj
        # but the general idea is that we want to process commands
        # from the main file right before starting up, so that we have a full
        # list of commands (otherwise, we may only get a partial list)

        # since we can pretty much guarantee that the ready function will only be
        # ran at the end for a variety of reasons, we can hook onto this
        # rather easily without many problems... unless someone's adding commands
        # after the bot has started, to which they should be using add_prefixed_command
        # anyways.
        def cursed_override(old_ready):
            async def new_ready():
                for _, func in inspect.getmembers(
                    sys.modules["__main__"],
                    predicate=lambda x: isinstance(x, MolterCommand)
                    and not x.is_subcommand(),
                ):
                    self.add_prefixed_command(func)
                await old_ready()

            return new_ready

        self.client._ready = cursed_override(self.client._ready)

        self.client.event(self._handle_prefixed_commands, name="on_message_create")  # type: ignore
        self.client.event(self.on_molter_command_error, name="on_molter_command_error")  # type: ignore

    def add_prefixed_command(self, command: MolterCommand) -> None:
        """
        Add a prefixed command to the client.

        Args:
            command (`MolterCommand`): The command to add.
        """
        if command.is_subcommand():
            raise ValueError(
                "You cannot add subcommands to the client - add the base command"
                " instead."
            )

        command._parse_parameters()

        if self.prefixed_commands.get(command.name):
            raise ValueError(
                "Duplicate command! Multiple commands share the name/alias:"
                f" {command.name}."
            )
        self.prefixed_commands[command.name] = command

        for alias in command.aliases:
            if self.prefixed_commands.get(alias):
                raise ValueError(
                    "Duplicate command! Multiple commands share the name/alias:"
                    f" {alias}."
                )
            self.prefixed_commands[alias] = command

    def remove_prefixed_command(self, name: str):
        """
        Removes a command if it exists.
        If an alias is specified, only the alias will be removed.

        Args:
            name (`str`): The command to remove.
        """
        command = self.prefixed_commands.pop(name, None)

        if command is None:
            return

        if name in command.aliases:
            command.aliases.remove(name)
            return

        for alias in command.aliases:
            self.prefixed_commands.pop(alias, None)

    def prefixed_command(
        self,
        name: typing.Optional[str] = None,
        *,
        aliases: typing.Optional[typing.List[str]] = None,
        help: typing.Optional[str] = None,
        brief: typing.Optional[str] = None,
        usage: typing.Optional[str] = None,
        enabled: bool = True,
        hidden: bool = False,
        ignore_extra: bool = True,
        type_to_converter: typing.Optional[
            typing.Dict[type, typing.Type[MolterConverter]]
        ] = None,
    ) -> typing.Callable[..., MolterCommand]:
        """
        WARNING:
            This method is deprecated and will be removed in a future release.
            Please use `molter.prefixed_command` instead.

        A decorator to declare a coroutine as a Molter prefixed command.

        Parameters:
            name (`str`, optional): The name of the command.
            Defaults to the name of the coroutine.

            aliases (`list[str]`, optional): The list of aliases the
            command can be invoked under.

            help (`str`, optional): The long help text for the command.
            Defaults to the docstring of the coroutine, if there is one.

            brief (`str`, optional): The short help text for the command.
            Defaults to the first line of the help text, if there is one.

            usage(`str`, optional): A string displaying how the command
            can be used. If no string is set, it will default to the
            command's signature. Useful for help commands.

            enabled (`bool`, optional): Whether this command can be run
            at all. Defaults to True.

            hidden (`bool`, optional): If `True`, the default help
            command (when it is added) does not show this in the help
            output. Defaults to False.

            ignore_extra (`bool`, optional): If `True`, ignores extraneous
            strings passed to a command if all its requirements are met
            (e.g. ?foo a b c when only expecting a and b).
            Otherwise, an error is raised. Defaults to True.

            type_to_converter (`dict[type, type[MolterConverter]]`, optional): A dict
            that associates converters for types. This allows you to use
            native type annotations without needing to use `typing.Annotated`.
            If this is not set, only interactions.py classes will be converted using
            built-in converters.

        Returns:
            `MolterCommand`: The command object.
        """

        def wrapper(func):
            logger.warning(
                "`MolterManager.prefixed_command` (commonly seen as"
                " `@molt.prefixed_command`) and its aliases has been deprecated. Use"
                " `molter.prefixed_command` or its aliases instead."
            )

            return MolterCommand(
                callback=func,
                name=name or func.__name__,
                aliases=aliases or [],
                help=help,
                brief=brief,
                usage=usage,  # type: ignore
                enabled=enabled,
                hidden=hidden,
                ignore_extra=ignore_extra,
                type_to_converter=type_to_converter  # type: ignore
                or getattr(func, "_type_to_converter", {}),
            )

        return wrapper

    prefix_command = prefixed_command
    text_based_command = prefixed_command

    async def generate_prefixes(
        self, client: interactions.Client, msg: interactions.Message
    ) -> typing.Union[str, typing.Iterable[str]]:
        """
        Generates a list of prefixes a prefixed command can have based on the client and message.
        This can be overwritten by passing a function to generate_prefixes on initialization.

        Args:
            client (`interactions.Client`): The client instance.
            msg (`interactions.Message`): The message sent.

        Returns:
            `str` | `Iterable[str]`: The prefix(es) to check for.
        """
        return self.default_prefix  # type: ignore

    async def on_molter_command_error(
        self, context: MolterContext, error: Exception
    ) -> None:
        """
        A function that is called when a molter command errors out.
        By default, this function outputs to the default logging place.

        Args:
            context (`MolterContext`): The context in which the error occured.
            error (`Exception`): The exception raised by the molter command.
        """

        out = traceback.format_exception(type(error), error, error.__traceback__)
        logger.error(
            "Ignoring exception in {}:{}{}".format(
                f"molter cmd / {context.invoked_name}",
                "\n" if len(out) > 1 else " ",
                "".join(out),
            ),
        )

    async def _create_context(self, msg: interactions.Message) -> MolterContext:
        """
        Creates a `MolterContext` object from the given message.

        Args:
            msg (`interactions.Message`): The message to create a context from.

        Returns:
            `MolterContext`: The context generated.
        """
        # weirdly enough, sometimes this isn't set right
        msg._client = self.client._http

        channel = await interactions.get(
            self.client, interactions.Channel, object_id=int(msg.channel_id)
        )

        if (guild_id := msg.guild_id) or (guild_id := channel.guild_id):
            guild = await utils._wrap_lib_exception(
                interactions.get(
                    self.client, interactions.Guild, object_id=int(guild_id)
                )
            )
        else:
            guild = None

        return MolterContext(  # type: ignore
            client=self.client,
            message=msg,
            user=msg.author,  # type: ignore
            member=msg.member,
            channel=channel,
            guild=guild,
        )

    async def _handle_prefixed_commands(self, msg: interactions.Message):
        """
        Determines if a command is being triggered and dispatch it.

        Args:
            msg (`interactions.Message`): The message created.
        """

        if not msg.content or msg.author.bot:
            return

        prefixes = await self.generate_prefixes(self.client, msg)

        if isinstance(prefixes, str):
            # its easier to treat everything as if it may be an iterable
            # rather than building a special case for this
            prefixes = (prefixes,)

        if prefix_used := next(
            (prefix for prefix in prefixes if msg.content.startswith(prefix)), None
        ):
            context = await self._create_context(msg)
            context.prefix = prefix_used
            context.content_parameters = utils.remove_prefix(msg.content, prefix_used)
            command = self.client.molter

            while True:
                first_word: str = utils.get_first_word(context.content_parameters)  # type: ignore
                if isinstance(command, MolterCommand):
                    new_command = command.subcommands.get(first_word)
                else:
                    new_command = command.prefixed_commands.get(first_word)
                if not new_command or not new_command.enabled:
                    break

                command = new_command
                context.content_parameters = utils.remove_prefix(
                    context.content_parameters, first_word
                ).strip()

                if command.subcommands and command.hierarchical_checking:
                    try:
                        await command._run_checks(context)
                    except Exception as e:
                        if command.error_callback:
                            await command.error_callback(context, e)  # type: ignore
                        elif command.extension and command.extension._error_callback:
                            await command.extension._error_callback(context, e)
                        else:
                            self.client._websocket._dispatch.dispatch(
                                "on_molter_command_error", context, e
                            )
                        return

            if isinstance(command, MolterManager):
                command = None

            if command and command.enabled:
                # this looks ugly, ik
                context.invoked_name = utils.remove_suffix(
                    utils.remove_prefix(msg.content, prefix_used),
                    context.content_parameters,
                ).strip()
                context.args = utils.get_args_from_str(context.content_parameters)
                context.command = command

                try:
                    self.client._websocket._dispatch.dispatch(
                        "on_molter_command", context
                    )
                    await command(context)
                except Exception as e:
                    if command.error_callback:
                        await command.error_callback(context, e)  # type: ignore
                    elif command.extension and command.extension._error_callback:
                        await command.extension._error_callback(context, e)
                    else:
                        self.client._websocket._dispatch.dispatch(
                            "on_molter_command_error", context, e
                        )
                finally:
                    self.client._websocket._dispatch.dispatch(
                        "on_molter_command_complete", context
                    )


def setup(
    client: interactions.Client,
    *,
    default_prefix: typing.Optional[typing.Union[str, typing.Iterable[str]]] = None,
    generate_prefixes: typing.Optional[
        typing.Callable[
            [interactions.Client, interactions.Message],
            typing.Coroutine[
                typing.Any, typing.Any, typing.Union[str, typing.Iterable[str]]
            ],
        ]
    ] = None,
    on_molter_command_error: typing.Optional[
        typing.Callable[[MolterContext, Exception], typing.Coroutine]
    ] = None,
    **kwargs,
) -> MolterManager:
    """
    Allows setup of Molter through normal extension loading.
    It is recommended to use this function directly, though.

    Parameters:
        client (`interactions.Client`): The client instance.
        default_prefix (`str | typing.Iterable[str]`, optional): \
            The default prefix to use. Defaults to None.
        generate_prefixes (`typing.Callable`, optional): An asynchronous function \
            that takes in a `Client` and `Message` object and returns either a \
            string or an iterable of strings. Defaults to None.
        on_molter_command_error (`typing.Callable`, optional): An asynchronous function \
            that takes in a `MolterContext` and `Exception` to handle errors that occur \
            when running molter commands. By default, molter will output the error to \
            the default logging place and ignore it. The error event can also be listened \
            to by listening to the "on_molter_command_error" event.

        If neither `default_prefix` or `generate_prefixes` are provided, the bot
        defaults to using it being mentioned as its prefix.

    Returns:
        `MolterManager`: The class that deals with all things Molter.
    """
    return MolterManager(
        client=client,
        default_prefix=default_prefix,
        generate_prefixes=generate_prefixes,
        on_molter_command_error=on_molter_command_error,
        **kwargs,
    )



import typing

import interactions
from . import utils
from .command import MCT
from .command import MolterCommand
from .errors import CheckFailure

if typing.TYPE_CHECKING:
    from .context import MolterContext


__all__ = ("check", "has_permissions", "has_guild_permissions", "guild_only", "dm_only")


def check(
    check: typing.Callable[
        ["MolterContext"], typing.Coroutine[typing.Any, typing.Any, bool]
    ]
) -> typing.Callable[..., MCT]:
    """
    Add a check to a command.

    Args:
        check: A coroutine as a check for this command.
    """

    def wrapper(coro: MCT) -> MCT:
        if isinstance(coro, MolterCommand):
            coro.checks.append(check)
            return coro
        if not hasattr(coro, "checks"):
            coro.__checks__ = []  # type: ignore
        coro.__checks__.append(check)  # type: ignore
        return coro

    return wrapper


def has_permissions(
    *permissions: interactions.Permissions,
) -> typing.Callable[..., MCT]:
    """
    A check to see if the author has permissions specified for the specific context.
    Considers guild ownership, roles, and channel overwrites.
    This does not take into account implicit permissions.
    Works for DMs.

    Args:
        permissions (`interactions.Permissions`): A list of permissions to check.
    """

    combined_permissions = interactions.Permissions(0)
    for perm in permissions:
        combined_permissions |= perm

    async def _permission_check(ctx: "MolterContext"):
        result = combined_permissions in ctx.author_permissions

        if not result:
            raise CheckFailure(
                ctx, "You do not have the proper permissions to use this command."
            )
        return result

    return check(_permission_check)  # type: ignore


def has_guild_permissions(
    *permissions: interactions.Permissions,
) -> typing.Callable[..., MCT]:
    """
    A check to see if the author has permissions specified for the guild.
    Considers guild ownership and roles.
    This does not take into account implicit permissions.
    Will fail in DMs.

    Args:
        permissions (`interactions.Permissions`): A list of permissions to check.
    """

    combined_permissions = interactions.Permissions(0)
    for perm in permissions:
        combined_permissions |= perm

    async def _permission_check(ctx: "MolterContext"):
        if not ctx.guild or isinstance(ctx.author, interactions.User):
            raise CheckFailure(ctx, "This command cannot be used in private messages.")

        result = combined_permissions in utils.guild_permissions(ctx.author, ctx.guild)

        if not result:
            raise CheckFailure(
                ctx, "You do not have the proper permissions to use this command."
            )
        return result

    return check(_permission_check)  # type: ignore


def guild_only() -> typing.Callable[..., MCT]:
    """A check to make the command only run in guilds."""

    async def _guild_check(ctx: "MolterContext"):
        if not ctx.guild_id:
            raise CheckFailure(ctx, "This command cannot be used in private messages.")
        return True

    return check(_guild_check)  # type: ignore


def dm_only() -> typing.Callable[..., MCT]:
    """A check to make the command only run in DMs."""

    async def _guild_check(ctx: "MolterContext"):
        if ctx.guild_id:
            raise CheckFailure(
                ctx, "This command can only be used in private messages."
            )
        return True

    return check(_guild_check)  # type: ignore



import typing as _typing

import attrs

import interactions
from . import utils

if _typing.TYPE_CHECKING:
    from .command import MolterCommand
    from .base import MolterInjectedClient

__all__ = ("MolterContext",)


@attrs.define()
class MolterContext:
    """
    A special 'Context' object for `molter`'s commands.
    This does not actually inherit from `interactions._Context`.
    """

    client: "MolterInjectedClient" = attrs.field()
    """The bot instance."""
    message: interactions.Message = attrs.field()
    """The message this represents."""
    channel: interactions.Channel = attrs.field()
    """The channel this message was sent through."""
    user: interactions.User = attrs.field()
    """The user who sent the message."""

    member: _typing.Optional[interactions.Member] = attrs.field(default=None)
    """The guild member who sent the message, if applicable."""
    guild: _typing.Optional[interactions.Guild] = attrs.field(default=None)
    """The guild this message was sent through, if applicable."""

    invoked_name: str = attrs.field(init=False, default=None)
    """The name/alias used to invoke the command."""
    content_parameters: str = attrs.field(init=False, default=None)
    """The message content without the prefix or command."""
    command: "MolterCommand" = attrs.field(init=False, default=None)
    """The command invoked."""
    args: _typing.List[_typing.Any] = attrs.field(init=False, factory=list)
    """The arguments used for this command."""
    prefix: str = attrs.field(default=None)
    """The prefix used for this command."""

    extras: _typing.Dict[_typing.Any, _typing.Any] = attrs.field(
        init=False, factory=dict, repr=False
    )
    """Extras used for this context. These can contain your own custom data."""

    _guild_permissions: _typing.Optional[interactions.Permissions] = attrs.field(
        init=False,
        default=None,
        repr=False,
    )
    _channel_permissions: _typing.Optional[interactions.Permissions] = attrs.field(
        init=False,
        default=None,
        repr=False,
    )

    def __attrs_post_init__(self) -> None:
        for inter_object in (
            self.message,
            self.member,
            self.channel,
            self.guild,
        ):
            if (
                not inter_object
                or inter_object is interactions.MISSING
                or "_client" not in inter_object.__slots__  # type: ignore
            ):
                continue
            inter_object._client = self._http

        if self.member:
            # discord doesn't provide this field normally with messages, but
            # we can easily add it here for convenience
            self.member.user = self.user

    @property
    def author(self) -> _typing.Union[interactions.Member, interactions.User]:
        """
        Either the member or user who sent the message. Prefers member,
        but defaults to user if the member does not exist.
        This is useful for getting a Discord user, regardless of if the
        message was from a guild or not.

        This is different from both the API and interactions.py in that
        this can be a member object. It follows the conventions of other
        Python Discord libraries.
        """
        return self.member or self.user

    @property
    def me(self) -> _typing.Union[interactions.Member, interactions.User, None]:
        """Returns the bot member or user, if cached."""
        if not self.client.me:
            return None

        if not self.guild:
            return interactions.get(
                self.client,
                interactions.User,
                object_id=int(self.client.me.id),
                force="cache",
            )

        return interactions.get(
            self.client,
            interactions.Member,
            parent_id=int(self.guild.id),
            object_id=int(self.client.me.id),
            force="cache",
        )

    @property
    def bot(self) -> interactions.Client:
        """An alias to `MolterContext.client`."""
        return self.client

    @property
    def channel_id(self) -> interactions.Snowflake:
        """Returns the channel ID where the message was sent."""
        return self.message.channel_id  # type: ignore

    @property
    def guild_id(self) -> _typing.Optional[interactions.Snowflake]:
        """Returns the guild ID where the message was sent, if applicable."""
        return self.message.guild_id

    @property
    def _http(self) -> interactions.HTTPClient:
        """Returns the HTTP client the client has."""
        return self.client._http

    @property
    def bot_permissions(self) -> _typing.Optional[interactions.Permissions]:
        """
        Returns the permissions the bot has for this context, if the bot user is cached.

        This factors in ownership, roles, and channel overwrites.
        This does not take into account implicit permissions.
        """
        return utils.permissions(self.me, self.channel, self.guild) if self.me else None

    @property
    def author_permissions(self) -> interactions.Permissions:
        """
        Returns the permissions the sender of this context has.

        This factors in ownership, roles, and channel overwrites.
        This does not take into account implicit permissions.
        """
        return utils.permissions(self.author, self.channel, self.guild)

    @property
    def _sendable_channel(self) -> interactions.Channel:
        """
        Gets the channel to send a message for.
        We don't exactly need a channel with fully correct attributes,
        so we can use a dummy channel here.
        """
        return self.channel or interactions.Channel(
            id=self.channel_id, type=0, _client=self._http  # type: ignore
        )

    def typing(self) -> utils.Typing:
        """
        A context manager to send a typing state to a given channel
        as long as long as the wrapped operation takes.

        Usage:

        ```python
        async with ctx.typing():
            # do stuff here
        ```
        """
        return utils.Typing(self._http, int(self.channel_id))

    async def fetch_me(self) -> _typing.Union[interactions.Member, interactions.User]:
        """
        Fetches the bot member or user.

        This is more reliable than `MolterContext.me` as it can
        fetch data from Discord if needed.
        """
        if not self.client.me:
            raise ValueError("You must be logged in to use this.")

        if self.guild:
            return await interactions.get(
                self.client,
                interactions.Member,
                parent_id=int(self.guild.id),
                object_id=int(self.client.me.id),
            )
        else:
            return await interactions.get(
                self.client,
                interactions.User,
                object_id=int(self.client.me.id),
            )

    async def fetch_bot_permissions(self) -> interactions.Permissions:
        """
        Fetches the permissions the bot has for this context.

        This factors in ownership, roles, and channel overwrites.
        This does not take into account implicit permissions.

        This is more reliable than `MolterContext.bot_permissions`
        as it can fetch data from Discord if needed.
        """

        me = await self.fetch_me()
        return utils.permissions(me, self.channel, self.guild)

    async def send(
        self,
        content: _typing.Optional[str] = interactions.MISSING,  # type: ignore
        *,
        tts: _typing.Optional[bool] = interactions.MISSING,  # type: ignore
        files: _typing.Optional[
            _typing.Union[interactions.File, _typing.List[interactions.File]]
        ] = interactions.MISSING,  # type: ignore
        embeds: _typing.Optional[
            _typing.Union["interactions.Embed", _typing.List["interactions.Embed"]]
        ] = interactions.MISSING,  # type: ignore
        allowed_mentions: _typing.Optional[
            "interactions.MessageInteraction"
        ] = interactions.MISSING,  # type: ignore
        components: _typing.Optional[
            _typing.Union[
                "interactions.ActionRow",  # type: ignore
                "interactions.Button",  # type: ignore
                "interactions.SelectMenu",  # type: ignore
                _typing.List["interactions.ActionRow"],  # type: ignore
                _typing.List["interactions.Button"],  # type: ignore
                _typing.List["interactions.SelectMenu"],  # type: ignore
            ]
        ] = interactions.MISSING,  # type: ignore
        **kwargs,
    ) -> "interactions.Message":  # type: ignore
        """
        Sends a message in the channel where the message came from.

        Args:
            content (`str`, optional): The contents of the message as a string
            or string-converted value.

            tts (`bool`, optional): Whether the message utilizes the text-to-speech
            Discord programme or not.

            files (`interactions.File | list[interactions.File]`, optional):
            A file or list of files to be attached to the message.

            embeds (`interactions.Embed | list[interactions.Embed]`, optional):
            An embed, or list of embeds for the message.

            allowed_mentions (`interactions.MessageInteraction`, optional):
            The message interactions/mention limits that the message can refer to.

            components (`interactions.ActionRow | interactions.Button |
            interactions.SelectMenu | list[interactions.ActionRow] |
            list[interactions.Button] | list[interactions.SelectMenu]`, optional):
            A component, or list of components for the message.

        Returns:
            `interactions.Message`: The sent message as an object.
        """

        return await self._sendable_channel.send(
            content,
            tts=tts,
            files=files,
            embeds=embeds,
            allowed_mentions=allowed_mentions,
            components=components,  # type: ignore
            **kwargs,
        )

    async def reply(
        self,
        content: _typing.Optional[str] = interactions.MISSING,  # type: ignore
        *,
        tts: _typing.Optional[bool] = interactions.MISSING,  # type: ignore
        files: _typing.Optional[
            _typing.Union[interactions.File, _typing.List[interactions.File]]
        ] = interactions.MISSING,  # type: ignore
        embeds: _typing.Optional[
            _typing.Union["interactions.Embed", _typing.List["interactions.Embed"]]
        ] = interactions.MISSING,  # type: ignore
        allowed_mentions: _typing.Optional[
            "interactions.MessageInteraction"
        ] = interactions.MISSING,  # type: ignore
        components: _typing.Optional[
            _typing.Union[
                "interactions.ActionRow",  # type: ignore
                "interactions.Button",  # type: ignore
                "interactions.SelectMenu",  # type: ignore
                _typing.List["interactions.ActionRow"],  # type: ignore
                _typing.List["interactions.Button"],  # type: ignore
                _typing.List["interactions.SelectMenu"],  # type: ignore
            ]
        ] = interactions.MISSING,  # type: ignore
        **kwargs,
    ) -> "interactions.Message":  # type: ignore
        """
        Sends a new message replying to the old.

        Args:
            content (`str`, optional): The contents of the message as a string
            or string-converted value.

            tts (`bool`, optional): Whether the message utilizes the text-to-speech
            Discord programme or not.

            files (`interactions.File | list[interactions.File]`, optional):
            A file or list of files to be attached to the message.

            embeds (`interactions.Embed | list[interactions.Embed]`, optional):
            An embed, or list of embeds for the message.

            allowed_mentions (`interactions.MessageInteraction`, optional):
            The message interactions/mention limits that the message can refer to.

            components (`interactions.ActionRow | interactions.Button |
            interactions.SelectMenu | list[interactions.ActionRow] |
            list[interactions.Button] | list[interactions.SelectMenu]`, optional):
            A component, or list of components for the message.

        Returns:
            `interactions.Message`: The sent message as an object.
        """

        return await self.message.reply(
            content,
            tts=tts,
            files=files,
            embeds=embeds,
            allowed_mentions=allowed_mentions,
            components=components,  # type: ignore
            **kwargs,
        )



import asyncio
import functools
import inspect
import re
import typing

import interactions
import interactions.api.error as inter_errors

if typing.TYPE_CHECKING:
    from .base import MolterExtension
    from .command import MolterCommand

__all__ = (
    "SnowflakeType",
    "OptionalSnowflakeType",
    "ALL_PERMISSIONS",
    "remove_prefix",
    "remove_suffix",
    "when_mentioned",
    "when_mentioned_or",
    "maybe_coroutine",
    "ARG_PARSE_REGEX",
    "MENTION_REGEX",
    "get_args_from_str",
    "get_first_word",
    "escape_mentions",
    "guild_permissions",
    "permissions",
    "Typing",
    "DeferredTyping",
)

# most of these come from naff
# thanks, polls!

SnowflakeType = typing.Union[interactions.Snowflake, int, str]
OptionalSnowflakeType = typing.Optional[SnowflakeType]


T = typing.TypeVar("T")

ALL_PERMISSIONS = interactions.Permissions(0)

for perm in interactions.Permissions:
    ALL_PERMISSIONS |= perm


async def _wrap_lib_exception(function: typing.Awaitable[T]) -> typing.Optional[T]:
    try:
        return await function
    except inter_errors.LibraryException:
        return None


def _wrap_recursive(cmd: "MolterCommand", ext: "MolterExtension"):
    cmd.extension = ext
    cmd.callback = functools.partial(cmd.callback, ext)

    if cmd.error_callback:
        cmd.error_callback = functools.partial(cmd.error_callback, ext)

    for subcommand in cmd.all_commands:
        new_sub = _wrap_recursive(subcommand, ext)

        names = [subcommand.name] + subcommand.aliases
        for name in names:
            cmd.subcommands[name] = new_sub

    return cmd


def remove_prefix(string: str, prefix: str) -> str:
    """
    Removes a prefix from a string if present.

    Args:
        string (`str`): The string to remove the prefix from.
        prefix (`str`): The prefix to remove.

    Returns:
        The string without the prefix.
    """
    return string[len(prefix) :] if string.startswith(prefix) else string[:]


def remove_suffix(string: str, suffix: str) -> str:
    """
    Removes a suffix from a string if present.

    Args:
        string (`str`): The string to remove the suffix from.
        suffix (`str`): The suffix to remove.

    Returns:
        The string without the suffix.
    """
    return string[: -len(suffix)] if string.endswith(suffix) else string[:]


async def when_mentioned(bot: interactions.Client, _) -> typing.List[str]:
    """
    Returns a list of the bot's mentions.

    Returns:
        A list of the bot's possible mentions.
    """
    return [f"<@{bot.me.id}> ", f"<@!{bot.me.id}> "]  # type: ignore


def when_mentioned_or(
    *prefixes: str,
) -> typing.Callable[
    [interactions.Client, typing.Any],
    typing.Coroutine[typing.Any, typing.Any, typing.List[str]],
]:
    """
    Returns a list of the bot's mentions plus whatever prefixes are provided.

    This is intended to be used with initializing molter. If you wish to use
    it in your own function, you will need to do something similar to:

    `await when_mentioned_or(*prefixes)(bot, msg)`

    Args:
        prefixes (`str`): Prefixes to include alongside mentions.

    Returns:
        A list of the bot's mentions plus whatever prefixes are provided.
    """

    async def new_mention(bot: interactions.Client, _):
        return (await when_mentioned(bot, _)) + list(prefixes)

    return new_mention


async def maybe_coroutine(func: typing.Callable, *args, **kwargs):
    """Allows running either a coroutine or a function."""
    if inspect.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    else:
        return func(*args, **kwargs)


_quotes = {
    '"': '"',
    "‘": "’",
    "‚": "‛",
    "“": "”",
    "„": "‟",
    "⹂": "⹂",
    "「": "」",
    "『": "』",
    "〝": "〞",
    "﹁": "﹂",
    "﹃": "﹄",
    "＂": "＂",
    "｢": "｣",
    "«": "»",
    "‹": "›",
    "《": "》",
    "〈": "〉",
}
_start_quotes = frozenset(_quotes.keys())

_pending_regex = r"(1.*2|[^\t\f\v ]+)"
_pending_regex = _pending_regex.replace("1", f"[{''.join(list(_quotes.keys()))}]")
_pending_regex = _pending_regex.replace("2", f"[{''.join(list(_quotes.values()))}]")

ARG_PARSE_REGEX = re.compile(_pending_regex)
MENTION_REGEX = re.compile(r"@(everyone|here|[!&]?[0-9]{17,20})")


def get_args_from_str(input: str) -> typing.List[str]:
    """
    Get arguments from an input string.

    Args:
        input (`str`): The string to process.
    Returns:
        A list of arguments.
    """
    return ARG_PARSE_REGEX.findall(input)


def get_first_word(text: str) -> typing.Optional[str]:
    """
    Get a the first word in a string, regardless of whitespace type.

    Args:
        text (`str`): The text to process.
    Returns:
        The first word, if found.
    """
    return split[0] if (split := text.split(maxsplit=1)) else None


def escape_mentions(content: str) -> str:
    """
    Escape mentions that could ping someone in a string.

    This does not escape channel mentions as they do not ping anybody.

    Args:
        content (`str`): The string to escape.
    Returns:
        The escaped string.
    """
    return MENTION_REGEX.sub("@\u200b\\1", content)


def guild_permissions(
    member: interactions.Member, guild: interactions.Guild
) -> interactions.Permissions:
    """
    Computes the guild (role-only) permissions for a member.
    This factors in ownership and the roles of the member.
    This does not take into account implicit permissions.

    This uses the pseudocode featured in Discord's own documentation about
    permission overwrites as a base.

    Returns:
        `interactions.Permissions`: The guild permissions for the member
        that sent the message.
    """

    if int(member.id) == guild.owner_id:
        return ALL_PERMISSIONS

    roles = guild.roles

    role_everyone = next(r for r in roles if r.id == guild.id)
    permissions = interactions.Permissions(int(role_everyone.permissions))

    if member.roles:
        member_roles = [r for r in roles if int(r.id) in member.roles]
    else:
        member_roles = []

    for role in member_roles:
        permissions |= interactions.Permissions(int(role.permissions))

    if (
        permissions & interactions.Permissions.ADMINISTRATOR
        == interactions.Permissions.ADMINISTRATOR
    ):
        return ALL_PERMISSIONS

    return permissions


def _compute_overwrites(
    member: interactions.Member,
    guild: interactions.Guild,
    channel: interactions.Channel,
    base_permissions: interactions.Permissions,
):
    """Calculates and adds in overwrites based on the guild permissions."""

    permissions = base_permissions

    if overwrites := channel.permission_overwrites:
        if overwrite_everyone := next(
            (o for o in overwrites if o.id == str(guild.id)), None
        ):
            permissions &= ~interactions.Permissions(int(overwrite_everyone.deny))
            permissions |= interactions.Permissions(int(overwrite_everyone.allow))

        allow = interactions.Permissions(0)
        deny = interactions.Permissions(0)

        for role_id in member.roles:
            if overwrite_role := next(
                (o for o in overwrites if o.id == str(role_id)), None
            ):
                allow |= interactions.Permissions(int(overwrite_role.allow))
                deny |= interactions.Permissions(int(overwrite_role.deny))

        permissions &= ~deny
        permissions |= allow

        if overwrite_member := next(
            (o for o in overwrites if o.id == str(member.id)), None
        ):
            permissions &= ~interactions.Permissions(int(overwrite_member.deny))
            permissions |= interactions.Permissions(int(overwrite_member.allow))

    return permissions


def permissions(
    member: typing.Union[interactions.Member, interactions.User],
    channel: interactions.Channel,
    guild: typing.Optional[interactions.Guild],
) -> interactions.Permissions:
    """
    Computes the permissions for the member specified.
    This factors in ownership, roles, and channel overwrites.
    This does not take into account implicit permissions.

    This uses the pseudocode featured in Discord's own documentation about
    permission overwrites as a base.

    Returns:
        `interactions.Permissions`: The permissions for the member
        that sent the message.
    """

    if not guild or isinstance(member, interactions.User):
        # basic text permissions, no tts, yes view channel
        return interactions.Permissions(0b1111100110001000000)

    base_permissions = guild_permissions(member, guild)  # type: ignore
    return _compute_overwrites(member, guild, channel, base_permissions)  # type: ignore


class Typing:
    """
    A context manager to send a typing state to a given channel
    as long as long as the wrapped operation takes.

    Args:
        http (`interactions.HTTPClient`): The HTTP client to use.
        channel_id (`int`): The ID of the channel to send the typing state to.
    """

    __slots__ = ("_http", "channel_id", "_stop", "task")

    def __init__(self, http: interactions.HTTPClient, channel_id: int) -> None:
        self._http = http
        self.channel_id = channel_id

        self._stop: bool = False
        self.task = None

    async def _typing_task(self) -> None:
        while not self._stop:
            await self._http.trigger_typing(self.channel_id)
            await asyncio.sleep(5)

    async def __aenter__(self) -> None:
        self.task = asyncio.create_task(self._typing_task())

    async def __aexit__(self, *_) -> None:
        self._stop = True
        self.task.cancel()  # type: ignore


class DeferredTyping:
    """
    A dummy context manager to defer an interaction and then do nothing.

    Args:
        interaction (`interactions.CommandContext`): The interaction to defer.
        ephemeral (`bool`, optional) Whether the response is hidden or not.
    """

    __slots__ = ("interaction", "ephemeral")

    def __init__(
        self, interaction: interactions.CommandContext, ephemeral: bool = False
    ) -> None:
        self.interaction = interaction
        self.ephemeral = ephemeral

    async def __aenter__(self) -> None:
        await self.interaction.defer(self.ephemeral)

    async def __aexit__(self, *_) -> None:
        pass
